﻿/* ControllableParallelFlow.cs
 * 
 * Copyright (C) 2018 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.HVAC.Circuit.ControllableFlowSolver
{
  /// <summary>抵抗調整によって流量制御可能な並列回路</summary>
  public class ControllableParallelFlow : IFlowControllableBranch
  {

    #region インスタンス変数・プロパティ

    /// <summary>抵抗調整によって流量制御可能な可変流量回路リスト</summary>
    private IFlowControllableBranch[] branches;

    /// <summary>往流路の抵抗リストを設定・取得する</summary>
    public double[] SupplyResistances { get; set; }

    /// <summary>還流路の抵抗リストを設定・取得する</summary>
    public double[] ReturnResistances { get; set; }

    /// <summary>回路全体の合成抵抗係数</summary>
    private double ttlResist = 0;

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="branches">抵抗調整によって流量制御可能な可変流量回路リスト</param>
    public ControllableParallelFlow(IFlowControllableBranch[] branches)
    {
      this.branches = branches;
      SupplyResistances = new double[branches.Length];
      ReturnResistances = new double[branches.Length];
    }

    #endregion

    #region IFlowControllableBranch実装

    /// <summary>合成抵抗が変化したか否か</summary>
    public bool HasTotalResistanceChanged
    {
      get
      {
        foreach (IFlowControllableBranch brch in branches)
          if (brch.HasTotalResistanceChanged) return true;
        return false;
      }
    }

    /// <summary>回路の合計流量[m3/s]を取得する</summary>
    public double TotalFlowRate { get; private set; }

    /// <summary>回路の合計設定流量[m3/s]を取得する</summary>
    /// <returns>回路の合計設定流量[m3/s]</returns>
    public double GetTotalFlowSetPoint()
    {
      double ttlFlw = 0;
      for (int i = 0; i < branches.Length; i++)
        ttlFlw += branches[i].GetTotalFlowSetPoint();
      return ttlFlw;
    }

    /// <summary>必要最低差圧[kPa]を取得する</summary>
    /// <returns>必要最低差圧[kPa]</returns>
    public double GetMinimumPressure()
    {
      double ttlFlw = GetTotalFlowSetPoint();
      double prlP = 0;
      double minP = 0;
      for (int i = 0; i < branches.Length; i++)
      {
        prlP += ttlFlw * ttlFlw
          * (SupplyResistances[i] + ReturnResistances[i]);
        minP = Math.Max(prlP + branches[i].GetMinimumPressure(), minP);
        ttlFlw -= branches[i].GetTotalFlowSetPoint();
      }
      return minP;
    }

    /// <summary>回路全体の合成抵抗[kPa/(m3/s)^2]を取得する</summary>
    /// <returns>回路全体の合成抵抗[kPa/(m3/s)^2]</returns>
    public double GetTotalResistance()
    {
      //抵抗係数に変更がなければ前回の計算結果を使用
      if (!HasTotalResistanceChanged) return ttlResist;

      //最遠方から順に合成する
      int indx = branches.Length - 1;
      ttlResist = branches[indx].GetTotalResistance() + SupplyResistances[indx] + ReturnResistances[indx];
      for (int i = branches.Length - 2; 0 <= i; i--)
      {
        ttlResist = 1 / Math.Pow(Math.Sqrt(1 / ttlResist) + Math.Sqrt(1 / branches[i].GetTotalResistance()), 2);
        ttlResist += SupplyResistances[i] + ReturnResistances[i];
      }
      return ttlResist;
    }

    /// <summary>差圧にもとづいて流量を調整する</summary>
    /// <param name="pressure">差圧[kPa]</param>
    public void ControlFlowRate(double pressure)
    {
      TotalFlowRate = Math.Sqrt(pressure / GetTotalResistance());
      double ttlFlw = TotalFlowRate;
      for (int i = 0; i < branches.Length; i++)
      {
        if (ttlFlw <= 0) pressure = 0;
        else pressure -= ttlFlw * ttlFlw
            * (SupplyResistances[i] + ReturnResistances[i]);
        branches[i].ControlFlowRate(pressure);
        ttlFlw -= branches[i].TotalFlowRate;
      }
    }

    #endregion

  }
}
